# @author Neil Spring

require 'socket'
require 'timeout'
require 'base64'
require 'resolv'
require 'scriptroute/packets'
require 'scriptroute/ally'
require 'scriptroute/rockettrace'
require 'scriptroute/nameify'

module Scriptroute

class ScriptrouteError < StandardError
end

# A ScriptrouteConnection is an object that wraps the TCP
# socket used to connect to a scriptroute daemon.  It likely
# need not be used directly, unless accessing low level
# scriptroute commands not wrapped by the general
# Scriptroute module
class ScriptrouteConnection
  Client_name = "%s(%d)" % [  defined?(Etc) ? Etc.getlogin : ENV['USER'], Process.uid ]
  # this version is designed to emulate the 0.4.14 version of srinterpreter.
  VERSION = '0.4.14'  
  def initialize
    begin
      timeout(5) do 
        @s = TCPSocket.new 'localhost', 3356
      end
      reply = ""
      reply = one_line_command( "interpret v%s %s\n" % [ VERSION, Client_name ]  )
      if reply =~ /proceed( Hz=(\d+))/ then
        # puts "negotiated."
      else
        puts "failed to negotiate: %s" % reply
      end
    rescue Errno::ECONNREFUSED => e
      $stderr.puts "Connection refused connecting to scriptrouted, ensure that it is running"
      raise e
    end
  end
  # @param t [String] the command to issue, newline optional
  # @return [String, nil] the one-line reply, or nil if the connection timed out or was terminated.
  def one_line_command(t)
    begin
      reply = nil
      timeout(2) do 
        issue_command t
        reply = @s.readline
      end
    rescue EOFError
    end
    return reply
  end
  # @return [Hash] the configuration of the scriptroute daemon
  def get_config
    ret = Hash.new
    timeout(5) do 
      issue_command "showconfig\n"
      l = "do/while"
      # showconfig ends reply with an empty line.
      while l !~ /^#done/  && l !~ /^$/ 
        l = @s.readline
        if l =~ /^(\S+)\s+=\s+(.+)$/ then
          ret[$1] = $2
        elsif l =~ /^(\S+)\s+=\s*$/ then
          ret[$1] = nil
        elsif l =~ /^$/ then
          # end. 
        else
          puts "unparsed config: %s" % l 
        end
      end 
    end
    ret
  end
  # @param t [String] the command to issue, newline optional
  def issue_command(t)
    # $stderr.puts "writing %s" % t
    if t[-1] == "\n" then
      @s.write t
    else
      @s.write "%s\n" % [ t ]
    end
  end
  # @return [String,nil] A one-line reply, or nil if the connection terminated.
  def get_reply
    begin
      @s.readline
    rescue EOFError
      nil
    end
  end
end

  # A DelayedPacket is for constructing trains in send_train.
  Struct.new("DelayedPacket", :delay, :packet)

  # TimedPacket is a time, packet tuple, with a tsc value thrown in in case its useful
  class TimedPacket
    # @return [Float,nil]
    attr_accessor :time
    # @return [Fixnum,nil] The output of rdtsc can be a more
    #   useful value in calculating rtt when NTP's adjustments
    #   via skew cause trouble.
    attr_accessor :tsc
    # @return [IP]
    attr_accessor :packet

    # @param time [Float,nil] Seconds since the epoch, or nil if we didn't see the packet leave due to pcap (happens)
    # @param tsc [Fixnum,nil] Value of the cycle counter (rdtsc) or nil if not supported
    # @param packet [IP] The packet received.
    def initialize(time, tsc, packet)
      raise ArgumentError, "no packet" unless packet
      raise ArgumentError, "packet of the wrong class" unless packet.is_a?(IP)
      @time = time
      @tsc = tsc
      @packet = packet
    end
  end

  # A ProbeResponse is a pair of a probe and its response.
  # Scriptroute is designed around the idea that a general
  # purpose engine can recognize the response to any probe,
  # and be in charge of doing so, so that measurement tools
  # need not have the rights to look at every packet.  
  # 
  # This design does limit somewhat, since probes that are 
  # capable of soliciting more than one response (e.g., via 
  # fragmentation) will not be managed properly.
  class ProbeResponse
    # @return [TimedPacket]
    attr_accessor :probe
    # @return [TimedPacket]
    attr_accessor :response
    # @return [Float,nil] Provides the apparent round trip time of this probe-response pair, or nil if either time is missing.
    def rtt
      if response and probe and probe.time then
        response.time - probe.time
      else
        nil
      end
    end
  end
  private

  # Connection pool in case the script requires more than one connection to the server (for concurrent tests).
  class ConnectionPool
    @@connections_cache = []
    # Connection pool mutex.
    @@connections_mutex = Mutex.new
    # Fetch from the connection pool or create a new connection
    # @return [ScriptrouteConnection]
    def ConnectionPool.get_idle_connection 
      @@connections_mutex.lock
      if @@connections_cache.empty? then
        @@connections_mutex.unlock
        return ScriptrouteConnection.new
      else
        ret = @@connections_cache.shift
        @@connections_mutex.unlock
        return ret
      end
    end
    # @param c [ScriptrouteConnection] the connection to return to the pool
    # @return [void]
    def ConnectionPool.return_idle_connection(c)
      @@connections_mutex.synchronize {
        @@connections_cache.push(c)
      }
    end
  end
  public
  # Take a block to be executed with a ScriptrouteConnection
  # from the pool.  This is the function to use, since it
  # manages the pooled connections explicitly.  A leak in
  # connections would be bad.
  # @yield [ScriptrouteConnection] the connection for this block.
  # @return the output of the block
  def Scriptroute::with_scriptroute_connection
    c = ConnectionPool.get_idle_connection
    r = yield c
    ConnectionPool.return_idle_connection(c)
    r
  end
  # Check if the daemon is running by making a connection or
  # checking the pool for an unused but working connection.
  # @return [Boolean]
  def Scriptroute::is_daemon_running?
    begin
      with_scriptroute_connection do |c|
      end
      return true
    rescue Errno::ECONNREFUSED
      return false
    end
  end
  # Exit failure if the scriptroute daemon does not appear
  # to be running.  This is useful at the beginning of
  # scripts that expect the daemon to run, to avoid adding
  # extra error checking later.
  # @return [true] if the script does ot exit because the daemon
  #   is not running.
  def Scriptroute::daemon_running_or_exit
    begin
      with_scriptroute_connection do |c|
      end
      return true
    rescue Errno::ECONNREFUSED
      puts "Ensure that the scriptroute daemon is running and try again"
      exit 1
    end
  end
  # Query for the version of the daemon; useful if there's a
  # feature only supported in a particular version.  @return
  # @return [String] the version string provided by of the running daemon
  def Scriptroute::DaemonVersion
    with_scriptroute_connection do |c|
      puts c.one_line_command("version\n")
    end
  end
  # @return [String] An IP address associated with the
  #   hostname.  Currently just invokes Resolv.getaddress.
  def Scriptroute::dns_lookup(name)
    Resolv.getaddress name
  end
  private 
  @@config = nil
  public
  # @return [Hash] the configuration of the running daemon,
  #  useful for checking rate limiting parameters or filters
  #  if a specially configured daemon is needed.
  def Scriptroute::DaemonConfig
    ret = nil
    unless @@config then
      with_scriptroute_connection do |c|
        @@config = c.get_config
      end
    end
    @@config
  end
  # no op for backward compatibility with the srinterpreter version
  # that had a dedicated packet instance
  # @param s [String] A marshaled packet.
  # @return [String] The input parameter, which is sufficient for this implementation of {Scriptroute::send_train}.
  def Scriptroute::pkt_from_string(s)
    s
  end
  # This is the nut.
  # @param train [Array<Struct::DelayedPacket,Array>] an array of Struct::DelayedPacket entries or arrays containing delay, packet pairs.
  # @return [Array<Scriptroute::ProbeResponse>] an array of Scriptroute::ProbeResponse entries, comprising a probe and response.
  def Scriptroute::send_train(train)
    ret = nil
    return [] if train.empty? # easy?
    with_scriptroute_connection do |c|
      # issue the send train command.
      i = 1;
      total_delay = 0
      train.map! { |boxcar|
        raise ArgumentError, 'nil entry' unless boxcar
        if boxcar.is_a?(Array) then
          Struct::DelayedPacket.new(*boxcar)
        else
          raise ArgumentError, 'not a boxcar or an array' unless boxcar.is_a?(Struct::DelayedPacket)
          boxcar
        end
      }
      train.each { |boxcar|
        # needs validation.
        delay = boxcar[:delay]
        packet = boxcar[:packet]
        raise "no packet" unless packet
        if packet.is_a?(Scriptroute::IP) then
          packet = packet.marshal
        end
        raise "packet is #{packet.class}, not a string" unless packet.is_a?(String)
        encoded_packet = Base64.strict_encode64(packet) # strict_encode means no line feeds added.
        c.issue_command "sendtrain %d/%d %f %d %s\n" % [ i, train.length, delay<0 ? 0 : delay, encoded_packet.length, encoded_packet ]
        i+=1
        total_delay += delay
      }

      # allocate space for the responses, array of probe response pairs, each of which will be a time/packet pair.
      ret = Array.new(i-1) { ProbeResponse.new }

      # collect, waiting maybe 5 minutes more than needed just in case the daemon hangs on us.
      begin
        timeout(total_delay + 300) do
          while l = c.get_reply and l !~ /^done/ and l =~ /\S+/ do
            if l =~ /^ERROR/ then
              # have to throw an exception, although individual packet errors are possible, that's
              # not communicated by the daemon.
              $stderr.puts l
              if l=~ /disjoint train \missing packet #(\d+)/ then
                ret[ $1.to_i ].probe = nil
                ret[ $1.to_i ].response = nil
              end
              raise ScriptrouteError, l
            else
              if l =~ /^(-?\d+\.\d+)\/(\d+) (\d+)([<>]) \d+ (\S+)$/ then
                tv_s = $1.to_f
                rdtsc = $2
                pk = $3.to_i - 1
                io = ($4 == '>') ? 0 : 1
                st = $5
                packet_string = Base64.strict_decode64(st)
                p = IP.creator(packet_string)
                tv_s = nil if tv_s < 0 # didn't see it leave.
                tp = TimedPacket.new(tv_s, rdtsc, p)
                if io == 0 then
                  ret[ pk ].probe = tp
                else
                  ret[ pk ].response = tp
                end
              else
                puts "unparsed send_train response: #{l}"
              end
            end
          end # while
        end # timeout
      rescue TimeoutError
        $stderr.puts "timed out parsing responses for send_train"
      end
    end
    return ret
  end
end

# Local Variables:
# compile-command: "rake test"
# End:
