# base class of command options.  all options must have tags 
# and a helpful description... or at least a description. 
class CommandoOpt
  attr_reader :tags, :description
  protected
  def initialize(tags, description)
    raise ArgumentError, "description must be a string (not #{description.class})" unless(description.is_a?(String)) 
    if(!tags.is_a?(Array)) then
      tags = [tags]
    end
    @tags = tags
    @description = description
  end
  public
  # @return [String] the help text for this option,
  #  comprising tags, default, and description
  def help
    " %20s %4s %-72s" % [ tags.join(", "), 
      self.string_default, 
      @description ]
  end
  # @note Modifies argument (removing parsed options)
  # @param [Array<String>] argv the arguments to parse.
  def seek(argv)
    argv.each_with_index { |a,i|
      tags.each { |t|
        if(a == t) then
          argv.delete_at(i) 
          if(takes_argument?) then
            set(argv[i])
            argv.delete_at(i) 
          else
            set(TRUE) # if closure, then the arg is ignored.
          end
        end
      }
    }
  end
end

# simple class for setting variables in command processing.
class CommandoVar < CommandoOpt
  # @return [Symbol] the name of the variable to set
  attr_reader :varname 
  # @return [Object] the default value of the variable
  attr_reader :default

  # @return [Boolean] whether the option takes a parameter, 
  #   based on whether the default is true or false.
  def takes_argument?
    # not really right, it should check the arity of the
    # closure if present.
    !(@default == true || @default == false)
  end
  # @return [String] the default argument, if any, or an
  #  empty string if no arguments are taken.
  def string_default
    if takes_argument? then 
      @default.to_s
    else
      ""
    end
  end

  # basic assignment, if it's a string, we quote it, else
  # the interpreter should deal with numbers, and complain
  # if something should have been quoted.  This operation
  # allows crazy s**t to happen, so don't use this in setuid code.
  # 
  # @param argument [String] the value that the variable
  #  should take, either from the default or from the
  #  command line.
  # @return [Fixnum,String] likely the argument as
  #   interpreted by eval.  Not likely to be useful.
  def set(argument)
    Kernel.eval( if(@default.is_a?(String)) then
                   "#{@varname} = \"#{argument}\""
                 else
                  "#{@varname} = #{argument}"
                 end
                )
  end

  # @param tags [Array<String>,String] the options to
  #   recognize, either just a string "--option" or an array [
  #   "-o", "--option" ]
  # @param description [String] help text to describe this option.
  # @param varname [Symbol] a global symbol to set to the
  #   value, e.g., ":$Option"
  # @param default [Object] a default value for the option,
  #   if not set, useful for the help description.  If the default
  #   is not a string, the assignment evals the variable unquoted.
  #   (That is, integer arguments should work as numbers without
  #   a need to convert.)
  def initialize(tags, description, varname, default=nil) 
    # should probably ensure that this isn't running setuid.
    super(tags, description)
    @varname = varname
    @default = default
    set(@default)
  end
end

# instead of setting a variable, execute an anonymous function.
# if the function takes an arg, gobble the next command line 
# argument.   If an argument is taken, the function is responsible
# for converting the string to whatever type is desired, and 
# validating the input -- raising an exception if a failure occurs.
class CommandoClosure < CommandoOpt
  attr_reader :closure
  # @return [Boolean] whether the option takes a parameter
  def takes_argument?
    (@closure.arity == -2) # means it takes one arg.
  end

  # @return [String] the default value, mostly for
  #  compatibility with CommandoVar.
  def string_default
    if takes_argument? then 
      "[x]" # return something to appease the help msg.
    else
      ""
    end
  end

  # basic assignment, if it's a string, we quote it, else
  # the interpreter should deal with numbers, and complain
  # if something should have been quoted.  This operation
  # allows crazy s**t to happen, so don't use this in setuid code.
  # 
  # @param argument [String] the value that the variable
  #  should take, either from the default or from the
  #  command line.
  def set(argument)
    case @closure.arity 
    when -1 
      @closure.call 
    when -2 
      @closure.call argv[i]
      argv.delete_at(i) 
    else
      raise "strange closure takes #{@closure.arity} args"
    end
  end

  # @param tags [Array<String>,String] the options to
  #   recognize, either just a string "--option" or an array [
  #   "-o", "--option" ]
  # @param description [String] help text to describe this option.
  # @param closure [Proc] the method to invoke if this option is 
  #   seen; the number of parameters to the Proc determines how 
  #   many subsequent arguments are considered parameters for this 
  #   option.
  def initialize(tags, description, closure)
    raise ArgumentError, "CommandoClosure takes a lambda as the third arg" unless( closure.is_a?(Proc) )
    # ruby1.8 seemed to use -1/-2 here; ruby1.9 seems to use
    # 0/1 here. accept em all.
    raise ArgumentError, "Closure for #{tags} should take zero or one args, not #{closure.arity}" unless(closure.arity == -1 or closure.arity == -2 or closure.arity == 0 or closure.arity == 1)
    
    super(tags, description)
    @closure = closure
  end
end

# @example
#   require "scriptroute/commando"
# 
#   c = Commando.new(ARGV,  # allows substitution by srclient.rb
#                    [ CommandoVar.new( "--start-speed",  
#                                      "Kbps rate to start" ,
#                                      :$StartSpeedKbps, 20),
#                      CommandoVar.new( "--train-length", 
#                                      "Filler packets in train" ,
#                                      :$TrainLength, 20 ),
#                      CommandoVar.new( "--hop",          
#                                      "Specific hop to study (ttl=hop and hop-1), 0 for e2e" ,
#                                      :$Hop, 0 ),
#                      CommandoVar.new( [ "--verbose", "-v" ], 
#                                      "print gobs of messages",
#                                      :$VERBOSE, false ) ],
#                    "destination-host")
# 
#   raise "must start with a positive rate" if($StartSpeedKbps <= 0)
#   if(ARGV[0] == nil) then
#     c.usage
#     exit
#   end
#
# Main class for option parsing.
#
# Limitations:
# doesn't currently grok the single letter getopt-style
# arguments, so "foo.rb -v -d -u" cannot be expressed as
# "foo.rb -vdu". 
#
class Commando
  # Print a usage message to stdout.
  # @return [void]
  def usage()
    puts "Usage: #{$0} [options] #{@after_options_name}"
    puts " %18s %4s %-72s" % [ "option", "default", "description" ]
    puts @options.map { |o|
      o.help
    }.join("\n")
  end
  # Parse any recognized command line options, removing
  #  anything parsed from ARGV.
  # @param argv [Array<String>] ARGV as provided by the interpreter.
  # @param options [Array [CommandoVar] an array of option descriptions
  # @param after_options_name [String] the description of
  #  what comes after the options parsed.
  def initialize(argv, options, after_options_name)
    @options = options
    @after_options_name = after_options_name
    helpopt = CommandoClosure.new(["--help", "-h"], "this help", 
                                  lambda { |a| self.usage; exit 0 })
    begin
      # helpopt goes last, in case the programmer wrote
      # a -h option in.
      (options + [helpopt]).each { |o| o.seek(argv) }
    rescue NameError => e 
      puts "Error in command line parsing: #{e}"
      puts 
      self.usage
      exit
    end
  end 
end
