#!/usr/bin/ruby
# generates a pretty version of an ip address using
# a reverse dns lookup and running the name through
# undns for decoding.

# one might modify this to consult other databases
# in the process.

require 'scriptroute'

$have_undns = begin
                require 'undns'
                true
              rescue LoadError, SecurityError
                false
              end
$have_origins = begin
                  $have_undns and Undns.origins_init
                  true
                rescue
                  false
                end
$have_socket = begin
                 require 'socket'
                 true
               rescue LoadError, SecurityError
                 $stderr.puts "couldn't load socket"
                 false
               end
$have_resolv = if(!$have_socket or Socket.gethostbyname("127.0.0.1")[0] !~ /localhost/) then
                 begin
                   # if socket library is not there or doesn't effectively do reverse lookups (dammit)
                   require 'resolv'
                   true
                 rescue LoadError, SecurityError
                   $stderr.puts "couldn't load resolv"
                   false
                 end
               else
                 false # don't need it.
               end

module Scriptroute
def Scriptroute::nameify(addr)
  unless addr.is_a?(Scriptroute::IPaddress) then
    # if not a known IP address, validate.
    if addr == nil or addr == '' then 
      # empty string. usually a mistake somewhere.
      $stderr.puts "nameify('') called from:\n  " + Kernel.caller.join("\n   ")
      return addr
    end
    if(! /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/.match(addr)) then
      # not an ip address looking thing.
      $stderr.puts "nameify(#{addr}) called from:\n  " + Kernel.caller.join("\n   ")
      return addr
    end
  end
  name =
    if($have_resolv) then
      begin
        Resolv.getnames(addr.to_s)[0]
      rescue ArgumentError => boom
        # seems to happen for RFC1918 space, but probably just for those
        # addresses without a name.
        $stderr.puts "argument error resolving '#{addr}' to name using resolv: " + boom
        nil 
      rescue SocketError => boom
        nil
      end
    elsif($have_socket) then
      begin
        name = (Socket.gethostbyname(addr.to_s)[0]).gsub(/\"/,'')
        name
      rescue ArgumentError => boom
        # seems to happen for RFC1918 space, but probably just for those
        # addresses without a name.
        $stderr.puts "argument error resolving '#{addr}' to name: " + boom
        nil 
      rescue SocketError => boom
        nil
      end
    else
      nil
    end
  attributes = if($have_undns) then
                 if(name) then
                   asn = Undns.get_asn_bydns(name)
                   if($have_origins) then
                     " [%d/%d]" % [ asn, Undns.origin_for_address_str(addr) ]
                   elsif(asn > 0) then
                     " [%d]" % [ asn ]
                   else
                     ""
                   end +
                     if(asn > 0) then
                       " {%s}" % [ Undns.get_loc(asn, name) ]
                     else
                       ""
                     end
                 elsif($have_origins) then
                   # don't have a name.
                   " [/%d]" % [ Undns.origin_for_address_str(addr) ]
                 else
                   ""
                 end
               else
                 ""
               end

  if(name) then
    name + ' (' + addr.to_s + ')' + attributes
  else
    addr.to_s + attributes
  end
end
end


if $0 == __FILE__ then
  # Undns.want_debug=1
  # Undns.want_conventions_debug=1 seems to fail to work for some reason.
  puts nameify("12.123.203.170")
  puts Undns.get_loc(7018, "tbr1-p012502.st6wa.ip.att.net")
  puts Undns.get_loc(0, "tbr1-p012502.st6wa.ip.att.net")
  puts Undns.get_asn_bydns("tbr1-p012502.st6wa.ip.att.net")
end

